#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../Set.hpp"

#include "../TreeUtils.hpp"

#define SET_NODE(Curr, Parent, Left, Right, Value) \
  (Curr)->parent = (Parent);                       \
  (Curr)->left = (Left);                           \
  (Curr)->right = (Right);                         \
  (Curr)->data.data = (Value);

SCENARIO ("SetIterators can be incremented", "[SetIterator][incr]")
{
  GIVEN ("An arbitrary Set")
  {
    using T = Counter<int>;
    TreeNode<T> dummy, a, b, c, d, e, f, g, h, i, j, k;
    SET_NODE (&dummy, &f, &k, &a, 0);
    SET_NODE (&a, &c, NULL, &b, 1);
    SET_NODE (&b, &a, NULL, NULL, 2);
    SET_NODE (&c, &f, &a, &d, 3);
    SET_NODE (&d, &c, NULL, &e, 4);
    SET_NODE (&e, &d, NULL, NULL, 5);
    SET_NODE (&f, &dummy, &c, &i, 6);
    SET_NODE (&g, &h, NULL, NULL, 7);
    SET_NODE (&h, &i, &g, NULL, 8);
    SET_NODE (&i, &f, &h, &k, 9);
    SET_NODE (&j, &k, NULL, NULL, 10);
    SET_NODE (&k, &i, &j, NULL, 11);

    SetIterator<T> end (&dummy), begin (&a), A (&a), B (&b), C (&c), D (&d),
      E (&e), F (&f), G (&g), H (&h), I (&i), J (&j), K (&k);
    THEN ("[2] incrementing iterators has them behave like successor")
    {
      Counters::reset ();
      REQUIRE (*++A == *B);
      REQUIRE (*++B == *C);
      REQUIRE (*++C == *D);
      REQUIRE (*++D == *E);
      REQUIRE (*++E == *F);
      REQUIRE (*++F == *G);
      REQUIRE (*++G == *H);
      REQUIRE (*++H == *I);
      REQUIRE (*++I == *J);
      REQUIRE (*++J == *K);
      REQUIRE (*++K == *end);
      REQUIRE (*++end == *begin);
      REQUIRE ((Counters::compares == 12 &&
                "we shouldn't perform any comparisons in increment"));
    }
  }
}

SCENARIO ("SetIterators can be decremented", "[SetIterator][decr]")
{
  GIVEN ("An arbitrary Set")
  {
    using T = Counter<int>;
    TreeNode<T> dummy, a, b, c, d, e, f, g, h, i, j, k;
    SET_NODE (&dummy, &f, &k, &a, 0);
    SET_NODE (&a, &c, NULL, &b, 1);
    SET_NODE (&b, &a, NULL, NULL, 2);
    SET_NODE (&c, &f, &a, &d, 3);
    SET_NODE (&d, &c, NULL, &e, 4);
    SET_NODE (&e, &d, NULL, NULL, 5);
    SET_NODE (&f, &dummy, &c, &i, 6);
    SET_NODE (&g, &h, NULL, NULL, 7);
    SET_NODE (&h, &i, &g, NULL, 8);
    SET_NODE (&i, &f, &h, &k, 9);
    SET_NODE (&j, &k, NULL, NULL, 10);
    SET_NODE (&k, &i, &j, NULL, 11);

    SetIterator<T> end (&dummy), begin (&a), A (&a), B (&b), C (&c), D (&d),
      E (&e), F (&f), G (&g), H (&h), I (&i), J (&j), K (&k);
    THEN ("[2] decrementing iterators has them behave like predecessor")
    {
      Counters::reset ();
      REQUIRE (*end == *--begin);
      REQUIRE (*K == *--end);
      REQUIRE (*J == *--K);
      REQUIRE (*I == *--J);
      REQUIRE (*H == *--I);
      REQUIRE (*G == *--H);
      REQUIRE (*F == *--G);
      REQUIRE (*E == *--F);
      REQUIRE (*D == *--E);
      REQUIRE (*C == *--D);
      REQUIRE (*B == *--C);
      REQUIRE (*A == *--B);
      REQUIRE ((Counters::compares == 12 &&
                "we shouldn't perform any comparisons in decrement"));
    }
  }
}

SCENARIO ("SetIterators can be compared", "[SetIterator][compared]")
{
  using T = Counter<int>;
  TreeNode<T> node1, node2;
  GIVEN ("Two nodes")
  {
    WHEN ("We have two iterators that refer to the same node")
    {
      SetIterator<T> iter1 (&node1), iter2 (&node1);
      THEN ("[1] They should evaluate to be equal when compared")
      {
        Counters::reset ();
        REQUIRE (iter1 == iter2);
        REQUIRE ((Counters::compares == 0 &&
                  "we shouldn't perform any comparisons on the data"));
      }
    }
    WHEN ("We have two iterators that refer to different nodes")
    {
      SetIterator<T> iter1 (&node1), iter2 (&node2);
      THEN ("[1] They should not be equal when compared")
      {
        Counters::reset ();
        REQUIRE (iter1 != iter2);
        REQUIRE ((Counters::compares == 0 &&
                  "we shouldn't perform any comparisons on the data"));
      }
    }
  }
}

template<typename T>
struct SetAccessor
{
  Tree<T> tree;
};

template<typename T>
TreeNode<T>*
root (Set<T> const& s)
{
  return reinterpret_cast<SetAccessor<T> const*> (&s)->tree.root ();
}

SCENARIO ("Set has begin()/end()", "[Set][begin-end]")
{
  GIVEN ("An empty Set")
  {
    Set<int> s;
    THEN ("[1] begin() should equal end()")
    {
      REQUIRE (s.begin () == s.end ());
    }
    WHEN ("It is const")
    {
      Set<int> const & sconst = s;
      THEN ("[1] begin() should equal end()")
      {
        REQUIRE (sconst.begin () == sconst.end ());
      }
    }
  }
  GIVEN ("An arbitrary Set")
  {
    Set<int> s;
    for (int val : {4, 2, 6, 1, 3, 5, 7})
    {
      s.insert (val);
    }

    THEN ("[1] iterators should work")
    {
      REQUIRE (s.begin() != s.end());
      REQUIRE (*s.begin() == 1);
      REQUIRE (*std::prev(s.end()) == 7);
      REQUIRE (std::prev(s.begin()) == s.end());
      REQUIRE (std::next(s.end()) == s.begin());

    }
    WHEN ("It is const")
    {
      Set<int> const & sconst = s;
      THEN ("[1] iterators should work")
      {
        REQUIRE (sconst.begin() != sconst.end());
        REQUIRE (*sconst.begin() == 1);
        REQUIRE (*std::prev(sconst.end()) == 7);
        REQUIRE (std::prev(sconst.begin()) == sconst.end());
        REQUIRE (std::next(sconst.end()) == sconst.begin());
      }
    }
  }
}

SCENARIO ("Range insertion works for Set", "[Set][range-insert]")
{
  GIVEN ("An empty range")
  {
    Set<int> aSet;
    std::vector<int> v;
    WHEN ("We range insert into a set")
    {
      aSet.insert (v.begin (), v.end ());
      THEN ("[1] nothing should happen")
      {
        REQUIRE (aSet.size () == 0);
        REQUIRE (aSet.begin () == aSet.end ());
      }
    }
  }
  GIVEN ("An single-element range")
  {
    Set<int> aSet, rangeSet;
    std::array<int, 1> arr{7};
    WHEN ("We range insert into a set")
    {
      aSet.insert (7);
      rangeSet.insert (arr.begin (), arr.end ());
      THEN ("[1] it behaves as if we called insert exactly once")
      {
        REQUIRE (aSet.size() == 1);
        REQUIRE (aSet.begin() != aSet.end());
        REQUIRE (as_tree (root (aSet)) == as_tree (root (rangeSet)));
      }
    }
  }
  GIVEN ("A multi-element range")
  {
    Set<int> aSet, rangeSet;
    aSet.insert (4);
    aSet.insert (2);
    aSet.insert (6);
    aSet.insert (1);
    aSet.insert (3);
    aSet.insert (5);
    aSet.insert (7);
    REQUIRE (aSet.size() == 7);
    REQUIRE (aSet.begin() != aSet.end());
    WHEN ("We range insert into a set")
    {
      std::array<int, 7> arr{4, 2, 6, 1, 3, 5, 7};
      rangeSet.insert (arr.begin (), arr.end ());
      THEN ("[2] it behaves as if we called insert on each element")
      {
        REQUIRE (aSet.size () == rangeSet.size ());
        REQUIRE (as_tree (root (aSet)) == as_tree (root (rangeSet)));
        REQUIRE (std::equal (aSet.begin (), aSet.end (), rangeSet.begin ()));
      }
    }
  }
}

SCENARIO ("Range erase works for Set", "[Set][range-erase]")
{
  GIVEN ("A set with some elements")
  {
    using T = Counter<char>;
    Set<T> a{T ('f'), T ('c'), T ('i'), T ('a'), T ('d'), T ('h'),
             T ('k'), T ('b'), T ('e'), T ('g'), T ('j')};
    REQUIRE (a.size() == 11);
    REQUIRE (a.begin() != a.end());
    WHEN ("We try to erase an empty range")
    {
      auto const old_string = as_tree (root (a));
      auto i = std::next (a.begin (), 5);
      auto result = a.erase (i, i);
      THEN ("[1] We do nothing and return an iterator to 'last'")
      {
        REQUIRE (old_string == as_tree (root (a)));
        REQUIRE (a.size () == 11);
        REQUIRE (i == result);
      }
    }

    WHEN ("We try to erase range of size one")
    {
      Set<T> range (a);
      Set<T> single (a);
      auto rem = std::next (range.begin (), 8);
      T removed = *rem;
      auto next = std::next (rem);
      auto i = std::next (single.begin (), 8);
      auto single_result = single.erase (i);
      auto range_result = range.erase (rem, next);
      THEN ("[1] we remove the single element and return an iterator to 'last'")
      {
        REQUIRE (removed.data + 1 == range_result->data);
        REQUIRE (*range_result == *single_result);
        REQUIRE (range_result == next);
        REQUIRE (as_tree (root (range)) == as_tree (root (single)));
      }
    }

    WHEN ("We try to erase a range of elements")
    {
      Set<T> range (a);
      Set<T> loop (range);
      auto rangeFirst = std::next (range.begin (), 1);
      auto rangeLast = std::prev (range.end (), 1);

      auto loopFirst = std::next (loop.begin (), 1);
      auto loopLast = std::prev (loop.end (), 1);

      auto rangeResult = a.erase (rangeFirst, rangeLast);
      auto loopResult = loopFirst;
      while (loopResult != loopLast)
      {
        loopResult = loop.erase (loopResult);
      }
      THEN (
        "[1] we remove the range of elements one-by-one and return an iterator "
        "to 'last'")
      {
        REQUIRE (*rangeResult == *loopResult);
        REQUIRE (rangeResult == rangeLast);
        REQUIRE (as_tree (root (range)) == as_tree (root (loop)));
      }
    }
    WHEN ("We try to erase the entire range of elements")
    {
      Set<T> empty;
      auto rangeResult = a.erase (a.begin (), a.end ());
      THEN ("[1] the set should be empty")
      {
        REQUIRE (rangeResult == a.end ());
        REQUIRE (a.size () == 0);
        REQUIRE (as_tree (root (a)) == as_tree (root (empty)));
      }
    }
  }
}

#undef SET_NODE
