#include "common.hpp"

#include "catch_config.hpp"

#include "catch.hpp"

#include "../TreeNode.hpp"

#define SET_NODE(Curr, Parent, Left, Right, Value) \
  (Curr)->parent = (Parent);                       \
  (Curr)->left = (Left);                           \
  (Curr)->right = (Right);                         \
  (Curr)->data.data = (Value);

SCENARIO ("TreeNode should have functioning constructors", "[TreeNode][ctor]")
{
  TreeNode<Counter<int>> left;
  TreeNode<Counter<int>> right;
  TreeNode<Counter<int>> parent;
  Counter<int> data;
  data.data = 4;

  WHEN ("we construct a TreeNode with a value and pointers")
  {
    Counters::reset ();
    TreeNode<Counter<int>> newNode (data, &parent, &left, &right);
    THEN ("[2] Everything was copied using a member intializer list")
    {
      REQUIRE (newNode.data == data);
      REQUIRE (newNode.parent == &parent);
      REQUIRE (newNode.left == &left);
      REQUIRE (newNode.right == &right);
      REQUIRE ((Counters::cctor == 1 && Counters::cassign == 0 &&
                "the copy constructor for data was called"));
    }
  }

  WHEN ("we copy construct a TreeNode")
  {
    TreeNode<Counter<int>> original;
    SET_NODE (&original, &parent, &left, &right, 4);
    Counters::reset ();

    TreeNode<Counter<int>> copy (original);
    THEN ("[2] Everything was copied using a member intializer list")
    {
      REQUIRE (original.data == copy.data);
      REQUIRE (original.parent == copy.parent);
      REQUIRE (original.left == copy.left);
      REQUIRE (original.right == copy.right);
      REQUIRE ((Counters::cctor == 1 && Counters::cassign == 0 &&
                "the copy constructor for data was called"));
    }
  }
}

SCENARIO ("TreeNodeUtils::is{Root,Leaf,LeftChild,RightChild}",
          "[TreeNodeUtils][query]")
{
  TreeNode<Counter<int>> dummy, n1, n2, n3;
  SET_NODE (&dummy, &n2, &n3, &n1, 0);
  SET_NODE (&n1, &n2, nullptr, nullptr, 1);
  SET_NODE (&n2, &dummy, &n1, &n3, 2);
  SET_NODE (&n3, &n2, nullptr, nullptr, 3);

  GIVEN ("A node which is not a root")
  {
    THEN ("[1] isRoot() reports it is not a root")
    {
      Counters::reset ();
      REQUIRE_FALSE (TreeNodeUtils::isRoot (&n1));
      REQUIRE_FALSE (TreeNodeUtils::isRoot (&n3));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is a root")
  {
    THEN ("[1] isRoot() reports it is a root")
    {
      Counters::reset ();
      REQUIRE (TreeNodeUtils::isRoot (&n2));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is not a leaf")
  {
    THEN ("[1] isLeaf() reports it is not a leaf")
    {
      Counters::reset ();
      REQUIRE_FALSE (TreeNodeUtils::isLeaf (&n2));
      REQUIRE_FALSE (TreeNodeUtils::isLeaf (&dummy));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is a leaf")
  {
    THEN ("[1] isLeaf() reports it is a leaf")
    {
      Counters::reset ();
      REQUIRE (TreeNodeUtils::isLeaf (&n1));
      REQUIRE (TreeNodeUtils::isLeaf (&n3));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is not a left child")
  {
    THEN ("[1] isLeftChild() reports it is not a left child")
    {
      Counters::reset ();
      REQUIRE_FALSE (TreeNodeUtils::isLeftChild (&n2));
      REQUIRE_FALSE (TreeNodeUtils::isLeftChild (&n3));
      REQUIRE_FALSE (TreeNodeUtils::isLeftChild (&dummy));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is a left child")
  {
    THEN ("[1] isLeftChild() reports it is a left child")
    {
      Counters::reset ();
      REQUIRE (TreeNodeUtils::isLeftChild (&n1));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is not a right child")
  {
    THEN ("[1] isRightChild() reports it is not a right child")
    {
      Counters::reset ();
      REQUIRE_FALSE (TreeNodeUtils::isRightChild (&n2));
      REQUIRE_FALSE (TreeNodeUtils::isRightChild (&n1));
      REQUIRE_FALSE (TreeNodeUtils::isRightChild (&dummy));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
  GIVEN ("A node which is a right child")
  {
    THEN ("[1] isRightChild() reports it is a right child")
    {
      Counters::reset ();
      REQUIRE (TreeNodeUtils::isRightChild (&n3));
      REQUIRE (Counters::all_zero ("No operators should have been called"));
    }
  }
}

SCENARIO ("TreeNodeUtils::{minimum,maximum} should work",
          "[TreeNodeUtils][min-max]")
{
  GIVEN ("Some arbitrary tree")
  {
    using Int = Counter<int>;
    TreeNode<Int> dummy, a, b, c, d, e, f, g;
    SET_NODE (&dummy, &a, &g, &d, 0);
    SET_NODE (&a, &dummy, &b, &c, 1);
    SET_NODE (&b, &a, &d, &e, 2);
    SET_NODE (&c, &a, &f, &g, 3);
    SET_NODE (&d, &b, nullptr, nullptr, 4);
    SET_NODE (&e, &b, nullptr, nullptr, 5);
    SET_NODE (&f, &c, nullptr, nullptr, 6);
    SET_NODE (&g, &c, nullptr, nullptr, 7);

    WHEN ("We get the minimum of a leaf")
    {
      Counters::reset ();
      THEN ("[1] we always return the same node")
      {
        REQUIRE (TreeNodeUtils::minimum (&d) == &d);
        REQUIRE (TreeNodeUtils::minimum (&e) == &e);
        REQUIRE (TreeNodeUtils::minimum (&f) == &f);
        REQUIRE (TreeNodeUtils::minimum (&g) == &g);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("We get the minimum of a tree")
    {
      Counters::reset ();
      THEN ("[2] we always traverse the left child")
      {
        REQUIRE (TreeNodeUtils::minimum (&a) == &d);
        REQUIRE (TreeNodeUtils::minimum (&b) == &d);
        REQUIRE (TreeNodeUtils::minimum (&c) == &f);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("We get the maximum of a leaf")
    {
      Counters::reset ();
      THEN ("[1] we always return the same node")
      {
        REQUIRE (TreeNodeUtils::maximum (&d) == &d);
        REQUIRE (TreeNodeUtils::maximum (&e) == &e);
        REQUIRE (TreeNodeUtils::maximum (&f) == &f);
        REQUIRE (TreeNodeUtils::maximum (&g) == &g);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("We get the maximum of a tree")
    {
      Counters::reset ();
      THEN ("[2] we always traverse the right child")
      {
        REQUIRE (TreeNodeUtils::maximum (&a) == &g);
        REQUIRE (TreeNodeUtils::maximum (&b) == &e);
        REQUIRE (TreeNodeUtils::maximum (&c) == &g);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
  }
}

SCENARIO ("TreeNodeUtils::successor should work", "[TreeNodeUtils][succ]")
{
  /*
                f
               / \
              /   \
             /     \
            /       \
           c         i
          / \       / \
         a   d     h   k
          \   \   /   /
           b   e g   j
   */
  GIVEN ("An arbitrary tree")
  {
    using T = Counter<int>;
    TreeNode<T> dummy, a, b, c, d, e, f, g, h, i, j, k;
    SET_NODE (&dummy, &f, &k, &a, 0);
    SET_NODE (&a, &c, NULL, &b, 1);
    SET_NODE (&b, &a, NULL, NULL, 2);
    SET_NODE (&c, &f, &a, &d, 3);
    SET_NODE (&d, &c, NULL, &e, 4);
    SET_NODE (&e, &d, NULL, NULL, 5);
    SET_NODE (&f, &dummy, &c, &i, 6);
    SET_NODE (&g, &h, NULL, NULL, 7);
    SET_NODE (&h, &i, &g, NULL, 8);
    SET_NODE (&i, &f, &h, &k, 9);
    SET_NODE (&j, &k, NULL, NULL, 10);
    SET_NODE (&k, &i, &j, NULL, 11);

    WHEN ("we get the successor of the root")
    {
      THEN ("[0.5] we always go to the min of the right subtree")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::successor (&f) == &g);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the successor of the header")
    {
      THEN ("[0.5] we always go to the min of the right subtree")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::successor (&dummy) == &a);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the successor to a node with a right")
    {
      THEN ("[0.5] we always go to the min of the right subtree")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::successor (&a) == &b);
        REQUIRE (TreeNodeUtils::successor (&c) == &d);
        REQUIRE (TreeNodeUtils::successor (&d) == &e);
        REQUIRE (TreeNodeUtils::successor (&i) == &j);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the successor to a left child with no right")
    {
      THEN ("[0.5] we always go to the parent")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::successor (&g) == &h);
        REQUIRE (TreeNodeUtils::successor (&h) == &i);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the successor to a right child with no right")
    {
      THEN (
        "[0.5] we traverse upward until we are not our parent's right -- and "
        "return the parent")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::successor (&b) == &c);
        REQUIRE (TreeNodeUtils::successor (&e) == &f);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the successor to the maximum node in our tree")
    {
      THEN (
        "[0.5] we traverse upward until we are not our parent's right -- and "
        "return the parent")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::successor (&k) == &dummy);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
  }
}

SCENARIO ("TreeNodeUtils::predecessor should work", "[TreeNodeUtils][pred]")
{
  /*
                f
               / \
              /   \
             /     \
            /       \
           c         i
          / \       / \
         a   d     h   k
          \   \   /   /
           b   e g   j
   */
  GIVEN ("An arbitrary tree")
  {
    using T = Counter<int>;
    TreeNode<T> dummy, a, b, c, d, e, f, g, h, i, j, k;
    SET_NODE (&dummy, &f, &k, &a, 0);
    SET_NODE (&a, &c, NULL, &b, 1);
    SET_NODE (&b, &a, NULL, NULL, 2);
    SET_NODE (&c, &f, &a, &d, 3);
    SET_NODE (&d, &c, NULL, &e, 4);
    SET_NODE (&e, &d, NULL, NULL, 5);
    SET_NODE (&f, &dummy, &c, &i, 6);
    SET_NODE (&g, &h, NULL, NULL, 7);
    SET_NODE (&h, &i, &g, NULL, 8);
    SET_NODE (&i, &f, &h, &k, 9);
    SET_NODE (&j, &k, NULL, NULL, 10);
    SET_NODE (&k, &i, &j, NULL, 11);

    WHEN ("we get the predecessor of the root")
    {
      THEN ("[0.5] we always go to the max of the left subtree")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::predecessor (&f) == &e);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the predecessor of the header")
    {
      THEN ("[0.5] we always go to the max of the left subtree")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::predecessor (&dummy) == &k);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the predecessor to a node with a left")
    {
      THEN ("[0.5] we always go to the max of the left subtree")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::predecessor (&k) == &j);
        REQUIRE (TreeNodeUtils::predecessor (&i) == &h);
        REQUIRE (TreeNodeUtils::predecessor (&h) == &g);
        REQUIRE (TreeNodeUtils::predecessor (&c) == &b);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the predecessor to a right child with no left")
    {
      THEN ("[0.5] we always go to the parent")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::predecessor (&b) == &a);
        REQUIRE (TreeNodeUtils::predecessor (&e) == &d);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the predecessor to a left child with no left")
    {
      THEN (
        "[0.5] we traverse upward until we are not our parent's left -- and "
        "return the parent")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::predecessor (&g) == &f);
        REQUIRE (TreeNodeUtils::predecessor (&j) == &i);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
    AND_WHEN ("we get the predecessor to the minimum node in our tree")
    {
      THEN (
        "[0.5] we traverse upward until we are not our parent's left -- and "
        "return the parent")
      {
        Counters::reset ();
        REQUIRE (TreeNodeUtils::predecessor (&a) == &dummy);
        REQUIRE (
          (Counters::compares == 0 && "we shouldn't perform any comparisons"));
      }
    }
  }
}

SCENARIO ("TreeNodeUtils::copy should copy a tree rooted at a node",
          "[TreeNodeUtils][copy]")
{
  GIVEN ("A nullptr")
  {
    TreeNode<Counter<int>>* dummy = nullptr;
    TreeNode<Counter<int>>* null = nullptr;
    Counters::reset ();
    WHEN ("We call TreeNodeUtils::copy()")
    {
      THEN ("[1] nothing should happen and nullptr should be returned")
      {
        REQUIRE (TreeNodeUtils::copy (null, dummy) == nullptr);
      }
    }
  }
  GIVEN ("A random tree")
  {
    /*
              d
             / \
            /   \
           b     f
          / \   / \
         a   c e   g
     */
    TreeNode<Counter<int>> dummy, a, b, c, d, e, f, g;
    SET_NODE (&dummy, &d, &b, &f, '\0');
    SET_NODE (&d, &dummy, &b, &f, 'd');
    SET_NODE (&b, &d, &a, &c, 'b');
    SET_NODE (&f, &d, &e, &g, 'f');
    SET_NODE (&a, &b, nullptr, nullptr, 'a');
    SET_NODE (&c, &b, nullptr, nullptr, 'c');
    SET_NODE (&e, &f, nullptr, nullptr, 'e');
    SET_NODE (&g, &f, nullptr, nullptr, 'g');
    TreeNode<Counter<int>> newRoot;
    Counters::reset ();
    WHEN ("We call TreeNodeUtils::copy()")
    {
      auto result = TreeNodeUtils::copy (&d, &newRoot);
      THEN ("[2] All nodes were replicated with left/right links")
      {
        REQUIRE ((Counters::ctor + Counters::cctor == 7 &&
                  "(copy) constructors were called the right number of times"));
        REQUIRE (result != nullptr);
        REQUIRE ((result->data == d.data && "root was copied"));
        REQUIRE ((result->left->data == b.data && "root->left was copied"));
        REQUIRE ((result->right->data == f.data && "root->right was copied"));
        REQUIRE (result->left->left->data == a.data);
        REQUIRE (result->left->right->data == c.data);
        REQUIRE (result->right->left->data == e.data);
        REQUIRE (result->right->right->data == g.data);
      }
      AND_THEN ("[1] All parent links were set")
      {
        REQUIRE (result != nullptr);
        REQUIRE (result->parent == &newRoot);
        REQUIRE (result->left->parent == result);
        REQUIRE (result->right->parent == result);
        REQUIRE (result->left->left->parent == result->left);
        REQUIRE (result->left->right->parent == result->left);
        REQUIRE (result->right->left->parent == result->right);
        REQUIRE (result->right->right->parent == result->right);
      }
    }
  }
}

SCENARIO ("TreeNodeUtils::clear should delete a tree rooted at a node",
          "[TreeNodeUtils][clear]")
{
  GIVEN ("A nullptr")
  {
    WHEN ("We call TreeNodeUtils::clear()")
    {
      Counters::reset ();
      TreeNode<Counter<int>>* null = nullptr;
      TreeNodeUtils::clear (null);
      THEN ("[1] Nothing happens")
      {
        REQUIRE (Counters::dtor == 0);
      }
    }
  }

  GIVEN ("An arbitrary tree")
  {
    auto dummy = new TreeNode<Counter<int>>;
    auto a = new TreeNode<Counter<int>>;
    auto b = new TreeNode<Counter<int>>;
    auto c = new TreeNode<Counter<int>>;
    auto d = new TreeNode<Counter<int>>;
    auto e = new TreeNode<Counter<int>>;
    auto f = new TreeNode<Counter<int>>;
    auto g = new TreeNode<Counter<int>>;
    SET_NODE (dummy, d, b, f, '\0');
    SET_NODE (d, dummy, b, f, 'd');
    SET_NODE (b, d, a, c, 'b');
    SET_NODE (f, d, e, g, 'f');
    SET_NODE (a, b, nullptr, nullptr, 'a');
    SET_NODE (c, b, nullptr, nullptr, 'c');
    SET_NODE (e, f, nullptr, nullptr, 'e');
    SET_NODE (g, f, nullptr, nullptr, 'g');
    Counters::reset ();
    WHEN ("We call TreeNodeUtils::clear()")
    {
      TreeNodeUtils::clear (d);
      THEN ("[3] All elements were destroyed")
      {
        REQUIRE (Counters::dtor == 7);
      }
    }
  }
}

#undef SET_NODE
