#ifndef COMMON_HPP_
#define COMMON_HPP_

#include <cstdlib>
#include <utility>

struct Counters
{
  static size_t ctor, cctor, mctor, cassign, massign, dtor, compares, vctor, cast;
  static inline void
  reset ()
  {
    ctor = cctor = mctor = cassign = massign = dtor = compares = vctor = cast = 0;
  }

  template <typename Info>
  static bool all_zero (Info&&)
  {
    return ctor == 0 && cctor == 0 && mctor == 0 && cassign == 0 && massign == 0 && dtor == 0 && compares == 0 && cast == 0 && vctor == 0;
  }
};

template<typename T>
struct Counter
{
  T data;

  constexpr explicit Counter (T const & v) noexcept
  : data (v)
  {
    ++Counters::vctor;
  }

  explicit operator T() const noexcept
  {
    ++Counters::cast;
    return data;
  }

  constexpr Counter () noexcept
  {
    ++Counters::ctor;
  }

  constexpr Counter (Counter const& o)  noexcept
  {
    data = o.data;
    ++Counters::cctor;
  }

  constexpr Counter (Counter&& o) noexcept
  {
    data = std::move (o.data);
    ++Counters::mctor;
  }

  ~Counter () noexcept
  {
    ++Counters::dtor;
  }

  constexpr Counter&
  operator= (Counter const& o) noexcept
  {
    if (&o != this)
    {
      data = o.data;
    }
    ++Counters::cassign;
    return *this;
  }

  constexpr Counter&
  operator= (Counter&& o) noexcept
  {
    if (&o != this)
    {
      data = std::move (o.data);
    }
    ++Counters::massign;
    return *this;
  }

  friend constexpr bool
  operator< (Counter const& a, Counter const& b) noexcept
  {
    ++Counters::compares;
    return a.data < b.data;
  }

  friend constexpr bool
  operator> (Counter const& a, Counter const& b) noexcept
  {
    ++Counters::compares;
    return a.data > b.data;
  }

  friend constexpr bool
  operator<= (Counter const& a, Counter const& b) noexcept
  {
    ++Counters::compares;
    return a.data <= b.data;
  }

  friend constexpr bool
  operator>= (Counter const& a, Counter const& b) noexcept
  {
    ++Counters::compares;
    return a.data >= b.data;
  }

  friend constexpr bool
  operator== (Counter const& a, Counter const& b) noexcept
  {
    ++Counters::compares;
    return a.data == b.data;
  }

  friend constexpr bool
  operator!= (Counter const& a, Counter const& b) noexcept
  {
    ++Counters::compares;
    return a.data != b.data;
  }

  friend std::ostream&
  operator<< (std::ostream& os, Counter const& a) noexcept
  {
    os << a.data;
    return os;
  }
};

#endif
